@@ -1,22 +1,22 @@
 /*
  * Pike Extension Modules - A collection of modules for the Pike Language
- * Copyright  2000-2003 The Caudium Group
- * 
+ * Copyright  2000-2005 The Caudium Group
+ *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
  * published by the Free Software Foundation; either version 2 of the
  * License, or (at your option) any later version.
- * 
+ *
  * This program is distributed in the hope that it will be useful, but
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  *
- * $Id: xml_sax.c,v 1.7 2003/03/31 14:07:01 grendel Exp $
+ * $Id: xml_sax.c,v 1.18 2005/01/03 17:08:07 kiwi Exp $
  */
 
 /*
@@ -61,12 +61,15 @@
 #define _GNU_SOURCE
 #endif
 
-#define SAX_DEBUG 1
-
-#include "global.h"
-RCSID("$Id: xml_sax.c,v 1.7 2003/03/31 14:07:01 grendel Exp $");
+//#define SAX_DEBUG 1
 
 #include "caudium_util.h"
+RCSID("$Id: xml_sax.c,v 1.18 2005/01/03 17:08:07 kiwi Exp $");
+
+#ifdef fp
+#undef fp
+#endif
+
 #include "xml_config.h"
 #include "xml_sax.h"
 
@@ -74,9 +77,10 @@ RCSID("$Id: xml_sax.c,v 1.7 2003/03/31 14:07:01 grendel Exp $");
 #include <string.h>
 #include <stdio.h>
 #include <stdarg.h>
+#include <stddef.h>
+#include <limits.h>
 
-#include <libxml/tree.h>
-#include <libxml/SAX.h>
+#include <libxml/SAX2.h>
 #include <libxml/entities.h>
 #include <libxml/parserInternals.h>
 
@@ -93,7 +97,8 @@ static unsigned int   __debug_indent;
 
 #define THIS ((sax_storage*)Pike_fp->current_storage)
 #define CB_ABSENT(_idx_) (THIS->callbackOffsets[(_idx_)] == -1)
-#define CB_API(_name_) {#_name_, 1}
+#define CB_API(_name_, _xmlAPI_) {#_name_, 1, offsetof(xmlSAXHandler,_xmlAPI_)}
+#define CB_OPT_API(_name_, _xmlAPI_) {#_name_, 0, offsetof(xmlSAXHandler,_xmlAPI_)}
 #define CB_CALL(_offset_, _args_) apply_low(THIS->callbacks, THIS->callbackOffsets[(_offset_)], (_args_))
 
 #define safe_push_text(_txt_) if ((_txt_)) push_text((_txt_)); else push_int(0)
@@ -123,7 +128,10 @@ static unsigned int   __debug_indent;
 #define getParameterEntitySAX 0x16
 #define cdataBlockSAX 0x17
 #define externalSubsetSAX 0x18
-#define CB_API_SIZE 0x19
+#define startElementNsSAX2 0x19
+#define endElementNsSAX2 0x20
+#define structuredError 0x21
+#define CB_API_SIZE 0x22
 
 typedef enum {
   PARSE_PUSH_PARSER = 0x01,
@@ -135,6 +143,7 @@ typedef struct
 {
   char   *name;
   int     req;
+  int     saxFunIdx;
 } pikeCallbackAPI;
 
 typedef struct 
@@ -179,39 +188,57 @@ static void pextsFatalError(void *ctx, const char *msg, ...);
 static xmlEntityPtr pextsGetParameterEntity(void *ctx, const xmlChar *name);
 static void pextsCdataBlock(void *ctx, const xmlChar *value, int len);
 static void pextsExternalSubset(void *ctx, const xmlChar *name, const xmlChar *externalID, const xmlChar *systemID);
-  
-static struct program  *sax_program;
+static void pextsStartElementNs(void * ctx, 
+                                const xmlChar * localname, 
+                                const xmlChar * prefix, 
+                                const xmlChar * URI, 
+                                int nb_namespaces, 
+                                const xmlChar ** namespaces, 
+                                int nb_attributes, 
+                                int nb_defaulted, 
+                                const xmlChar ** attributes);
+static void pextsEndElementNs(void * ctx, 
+                              const xmlChar * localname, 
+                              const xmlChar * prefix, 
+                              const xmlChar * URI);
+static void pextsStructuredError(void * userData, xmlErrorPtr error);
+
+static struct program  *xml_program;
+static struct program  *html_program;
 
 static pikeCallbackAPI  callback_api[] =
 {
-  CB_API(internalSubsetSAX),
-  CB_API(isStandaloneSAX),
-  CB_API(hasInternalSubsetSAX),
-  CB_API(hasExternalSubsetSAX),
-  CB_API(getEntitySAX),
-  CB_API(entityDeclSAX),
-  CB_API(notationDeclSAX),
-  CB_API(attributeDeclSAX),
-  CB_API(elementDeclSAX),
-  CB_API(unparsedEntityDeclSAX),
-  CB_API(startDocumentSAX),
-  CB_API(endDocumentSAX),
-  CB_API(startElementSAX),
-  CB_API(endElementSAX),
-  CB_API(referenceSAX),
-  CB_API(charactersSAX),
-  CB_API(ignorableWhitespaceSAX),
-  CB_API(processingInstructionSAX),
-  CB_API(commentSAX),
-  CB_API(warningSAX),
-  CB_API(errorSAX),
-  CB_API(fatalErrorSAX),
-  CB_API(getParameterEntitySAX),
-  CB_API(cdataBlockSAX),
-  CB_API(externalSubsetSAX),
+  CB_OPT_API(internalSubsetSAX, internalSubset),
+  CB_OPT_API(isStandaloneSAX, isStandalone),
+  CB_OPT_API(hasInternalSubsetSAX, hasInternalSubset),
+  CB_OPT_API(hasExternalSubsetSAX, hasExternalSubset),
+  CB_OPT_API(getEntitySAX, getEntity),
+  CB_OPT_API(entityDeclSAX, entityDecl),
+  CB_OPT_API(notationDeclSAX, notationDecl),
+  CB_OPT_API(attributeDeclSAX, attributeDecl),
+  CB_OPT_API(elementDeclSAX, elementDecl),
+  CB_OPT_API(unparsedEntityDeclSAX, unparsedEntityDecl),
+  CB_API(startDocumentSAX, startDocument),
+  CB_API(endDocumentSAX, endDocument),
+  CB_API(startElementSAX, startElement),
+  CB_API(endElementSAX, endElement),
+  CB_OPT_API(referenceSAX, reference),
+  CB_OPT_API(charactersSAX, characters),
+  CB_OPT_API(ignorableWhitespaceSAX, ignorableWhitespace),
+  CB_OPT_API(processingInstructionSAX, processingInstruction),
+  CB_OPT_API(commentSAX, comment),
+  CB_OPT_API(warningSAX, warning),
+  CB_OPT_API(errorSAX, error),
+  CB_OPT_API(fatalErrorSAX, fatalError),
+  CB_OPT_API(getParameterEntitySAX, getParameterEntity),
+  CB_OPT_API(cdataBlockSAX, cdataBlock),
+  CB_OPT_API(externalSubsetSAX, externalSubset),
+  CB_OPT_API(startElementNsSAX2, startElementNs),
+  CB_OPT_API(endElementNsSAX2, endElementNs),
+  CB_OPT_API(structuredError, serror),
 };
 
-static xmlSAXHandler   pextsSAX = {
+static struct _xmlSAXHandler   pextsSAX = {
   .internalSubset = pextsInternalSubset,
   .isStandalone = pextsIsStandalone,
   .hasInternalSubset = pextsHasInternalSubset,
@@ -238,7 +265,10 @@ static xmlSAXHandler   pextsSAX = {
   .fatalError = pextsFatalError,
   .getParameterEntity = pextsGetParameterEntity,
   .cdataBlock = pextsCdataBlock,
-  .externalSubset = pextsExternalSubset
+  .externalSubset = pextsExternalSubset,
+  .startElementNs = pextsStartElementNs,
+  .endElementNs = pextsEndElementNs,
+  .serror = pextsStructuredError,
 };
 
 static struct pike_string  *econtent_type;
@@ -247,6 +277,17 @@ static struct pike_string  *econtent_name;
 static struct pike_string  *econtent_prefix;
 static struct pike_string  *econtent_c1;
 static struct pike_string  *econtent_c2;
+static struct pike_string  *seDomain;
+static struct pike_string  *seCode;
+static struct pike_string  *seMessage;
+static struct pike_string  *seLevel;
+static struct pike_string  *seFile;
+static struct pike_string  *seLine;
+static struct pike_string  *seStr1;
+static struct pike_string  *seStr2;
+static struct pike_string  *seStr3;
+static struct pike_string  *seInt1;
+static struct pike_string  *seInt2;
 
 /* Parser callbacks */
 
@@ -266,7 +307,7 @@ static void pextsInternalSubset(void *ctx, const xmlChar *name, const xmlChar *e
   safe_push_text(systemID);
   push_svalue(&THIS->user_data);
   
-  CB_CALL(endElementSAX, 5);
+  CB_CALL(internalSubsetSAX, 5);
   pop_stack();
   
   DBG_FUNC_LEAVE();
@@ -344,8 +385,6 @@ static xmlEntityPtr pextsGetEntity(void *ctx, const xmlChar *name)
     return NULL;
   }
 
-  THIS->ctxt = (xmlParserCtxtPtr)ctx;
-  
   DBG_FUNC_LEAVE();
   return NULL;
 }
@@ -439,7 +478,7 @@ static void pextsAttributeDecl(void *ctx, const xmlChar *elem, const xmlChar *fu
   DBG_FUNC_LEAVE();
 }
 
-inline static struct mapping *tree2mapping(xmlElementContentPtr content)
+static struct mapping *tree2mapping(xmlElementContentPtr content)
 {
   struct mapping *ret;
   struct svalue   sv;
@@ -584,8 +623,10 @@ static void pextsStartElement(void *ctx, const xmlChar *name, const xmlChar **at
     npairs = 0;
     tmp = atts;
     while (tmp && *tmp) {
-      safe_push_text(*tmp++);
-      safe_push_text(*tmp++);
+      safe_push_text(*tmp);
+      tmp++;
+      safe_push_text(*tmp);
+      tmp++;
       npairs += 2;
     }
     f_aggregate_mapping(npairs);
@@ -650,7 +691,7 @@ static void pextsCharacters(void *ctx, const xmlChar *ch, int len)
   
   push_object(this_object());
   if (ch && len)
-    push_text(make_shared_binary_string((const char*)ch, len));
+    push_string(make_shared_binary_string((const char*)ch, (size_t) len));
   else
     push_int(0);
   push_svalue(&THIS->user_data);
@@ -672,7 +713,7 @@ static void pextsIgnorableWhitespace(void *ctx, const xmlChar *ch, int len)
   
   push_object(this_object());
   if (ch && len)
-    push_text(make_shared_binary_string((const char*)ch, len));
+    push_string(make_shared_binary_string((const char*)ch, (size_t) len));
   else
     push_int(0);
   push_svalue(&THIS->user_data);
@@ -838,7 +879,7 @@ static void pextsCdataBlock(void *ctx, const xmlChar *value, int len)
   
   push_object(this_object());
   if (value)
-    push_text(make_shared_binary_string((const char*)value, len));
+    push_string(make_shared_binary_string((const char*)value, (size_t) len));
   else
     push_int(0);
   push_svalue(&THIS->user_data);
@@ -869,42 +910,150 @@ static void pextsExternalSubset(void *ctx, const xmlChar *name, const xmlChar *e
   DBG_FUNC_LEAVE();
 }
 
-/*
- * We require the callbacks object to have all the methods found in the
- * xmlSAXHandler structure to be present in the Pike object passed as the
- * callbacks:
- *
- *   internalSubsetSAX
- *   isStandaloneSAX
- *   hasInternalSubsetSAX
- *   hasExternalSubsetSAX
- *   getEntitySAX
- *   entityDeclSAX
- *   notationDeclSAX
- *   attributeDeclSAX
- *   elementDeclSAX
- *   unparsedEntityDeclSAX
- *   setDocumentLocatorSAX
- *   startDocumentSAX
- *   endDocumentSAX
- *   startElementSAX
- *   endElementSAX
- *   referenceSAX
- *   charactersSAX
- *   ignorableWhitespaceSAX
- *   processingInstructionSAX
- *   commentSAX
- *   warningSAX
- *   errorSAX
- *   fatalErrorSAX
- *   getParameterEntitySAX
- *   cdataBlockSAX
- *   externalSubsetSAX
+static void pextsStartElementNs(void * ctx, 
+                                const xmlChar * localname, 
+                                const xmlChar * prefix, 
+                                const xmlChar * URI, 
+                                int nb_namespaces, 
+                                const xmlChar ** namespaces, 
+                                int nb_attributes, 
+                                int nb_defaulted, 
+                                const xmlChar ** attributes)
+{
+  int i;
+  
+  DBG_FUNC_ENTER();
+  if (CB_ABSENT(startElementNsSAX2)) {
+    DBG_FUNC_LEAVE();
+    return;
+  }
+
+  THIS->ctxt = (xmlParserCtxtPtr)ctx;
+  
+  push_object(this_object());
+  safe_push_text(localname);
+  safe_push_text(prefix);
+  safe_push_text(URI);
+
+  if (nb_namespaces) {
+    struct array     *arr;
+    
+    for (i = 0; i < nb_namespaces; i++)
+      safe_push_text(namespaces[i]);
+    arr = aggregate_array(nb_namespaces);
+    push_array(arr);
+  } else
+    push_int(0);
+
+  if (nb_attributes) {
+    struct array     *arr;
+    
+    push_int(nb_defaulted);
+    for (i = 0; i < nb_attributes; i++)
+     safe_push_text(attributes[i]);
+    arr = aggregate_array(nb_attributes);
+    push_array(arr);
+  } else {
+    push_int(0);
+    push_int(0);
+  }
+  push_svalue(&THIS->user_data);
+  CB_CALL(startElementNsSAX2, 7);
+  pop_stack();
+
+  DBG_FUNC_LEAVE();
+}
+
+static void pextsEndElementNs(void * ctx, 
+                              const xmlChar * localname, 
+                              const xmlChar * prefix, 
+                              const xmlChar * URI)
+{
+  DBG_FUNC_ENTER();
+  if (CB_ABSENT(endElementNsSAX2)) {
+    DBG_FUNC_LEAVE();
+    return;
+  }
+
+  THIS->ctxt = (xmlParserCtxtPtr)ctx;
+  
+  push_object(this_object());
+  safe_push_text(localname);
+  safe_push_text(prefix);
+  safe_push_text(URI);
+  push_svalue(&THIS->user_data);
+  CB_CALL(endElementNsSAX2, 7);
+  pop_stack();
+
+  DBG_FUNC_LEAVE();
+}
+
+static void pextsStructuredError(void * userData, xmlErrorPtr error)
+{
+  int mcount = 6;
+  
+  DBG_FUNC_ENTER();
+  if (CB_ABSENT(structuredError)) {
+    DBG_FUNC_LEAVE();
+    return;
+  }
+
+  push_object(this_object());
+
+  push_string(seDomain);
+  push_int(error->domain);
+  push_string(seCode);
+  push_int(error->code);
+  if (error->message) {
+    push_string(seMessage);
+    safe_push_text(error->message);
+    mcount++;
+  }
+  push_string(seLevel);
+  push_int(error->level);
+  if (error->file) {
+    push_string(seFile);
+    safe_push_text(error->file);
+    mcount++;
+  }
+  push_string(seLine);
+  push_int(error->line);
+  if (error->str1) {  
+    push_string(seStr1);
+    safe_push_text(error->str1);
+    mcount++;
+  }
+  if (error->str2) {  
+    push_string(seStr2);
+    safe_push_text(error->str2);
+    mcount++;
+  }
+  if (error->str3) {  
+    push_string(seStr3);
+    safe_push_text(error->str3);
+    mcount++;
+  }
+  push_string(seInt1);
+  push_int(error->int1);
+  push_string(seInt2);
+  push_int(error->int2);
+  f_aggregate_mapping(mcount << 1);
+  
+  push_svalue(&THIS->user_data);
+  CB_CALL(structuredError, 3);
+  pop_stack();
+
+  DBG_FUNC_LEAVE();
+}
+
+/* To learn which callbacks are required take a look at the callbacks_api
+ * array above.
  */
 static int is_callback_ok(struct object *callbacks, char **missing_method)
 {
   int                 ioff, i;
   struct identifier  *ident;
+  void              **tmp;
   
   if (!callbacks)
     return 0;
@@ -937,8 +1086,14 @@ static int is_callback_ok(struct object *callbacks, char **missing_method)
         *missing_method = callback_api[i].name;
       return 0;
     }
-    
+
+    /* Put the offset in the callbacks array and initialize the SAX handler
+     * structure with appropriate function pointer. The pointer arithmetic
+     * is hairy, but it works :>
+     */
     THIS->callbackOffsets[i] = ioff;
+    tmp = (void**)((char*)THIS->sax + callback_api[i].saxFunIdx);
+    *tmp = *((void**)((char*)(&pextsSAX) + callback_api[i].saxFunIdx));
     i++;
   }
   
@@ -954,7 +1109,7 @@ static void f_getLineNumber(INT32 args)
   if (!THIS->ctxt)
     push_int(-1);
   else
-    push_int(getLineNumber(THIS->ctxt));
+    push_int(xmlSAX2GetLineNumber(THIS->ctxt));
 }
 
 /*! @decl int getColumnNumber()
@@ -966,7 +1121,7 @@ static void f_getColumnNumber(INT32 args)
   if (!THIS->ctxt)
     push_int(-1);
   else
-    push_int(getColumnNumber(THIS->ctxt));
+    push_int(xmlSAX2GetColumnNumber(THIS->ctxt));
 }
 
 /*! @decl void create(string|object input, object callbacks, mapping|void  entities, mixed|void user_data, int|void input_is_data)
@@ -1018,18 +1173,23 @@ static void f_create(INT32 args)
       default:
         Pike_error("Incorrect number of arguments: expected between 2 and 4\n");
   }
-
   
   /* check whether file_obj is Stdio.File or derived */
   if (file_obj && find_identifier("read", file_obj->prog) < 0)
     Pike_error("Passed file object is not Stdio.File or derived from it\n");
+
+  /* The parser state is initialized so that no time is wasted for
+   * callbacks that aren't used by the calling Pike code.
+   */
+  THIS->sax = (xmlSAXHandler*)calloc(1, sizeof(xmlSAXHandler));
+  if (!THIS->sax)
+    SIMPLE_OUT_OF_MEMORY_ERROR("create", sizeof(xmlSAXHandler));
   
   /* check whether the callbacks object contains all the required methods
    * */
   if (!is_callback_ok(callbacks, &missing_method)) 
     Pike_error("Passed callbacks object is not valid. The %s method is missing.\n",
                missing_method);
-
   
   /* choose the parsing method */
   if (file_obj)
@@ -1044,7 +1204,6 @@ static void f_create(INT32 args)
   pop_n_elems(args);
   
   /* initialize the parser and state */
-  THIS->sax = &pextsSAX;
   switch (THIS->parsing_method) {
       case PARSE_PUSH_PARSER:
         THIS->file_obj = file_obj;
@@ -1068,14 +1227,13 @@ static void f_create(INT32 args)
   }
 }
 
-static void f_parse(INT32 args)
+static void f_parse_xml(INT32 args)
 {
   xmlDocPtr   doc = NULL;
   
   switch (THIS->parsing_method) {
-      case PARSE_PUSH_PARSER: {
-        break;
-      }
+      case PARSE_PUSH_PARSER:
+        Pike_error("Push parser not implemented yet. Please bug grendel@caudium.net to implement it.");
         
       case PARSE_MEMORY_PARSER:
         doc = xmlSAXParseMemory(THIS->sax, THIS->input_data->str, THIS->input_data->len, 1);
@@ -1085,6 +1243,30 @@ static void f_parse(INT32 args)
         doc = xmlSAXParseFileWithData(THIS->sax, THIS->input_data->str, 1, NULL);
         break;
   }
+  if ( doc != NULL )
+    xmlFreeDoc(doc);
+
+  push_int(0);
+}
+
+static void f_parse_html(INT32 args)
+{
+  xmlDocPtr   doc = NULL;
+  
+  switch (THIS->parsing_method) {
+      case PARSE_PUSH_PARSER:
+        Pike_error("Push parser not implemented yet. Please bug grendel@caudium.net to implement it.");
+        
+      case PARSE_MEMORY_PARSER:
+        doc = htmlSAXParseDoc(THIS->input_data->str, NULL, THIS->sax, NULL);
+        break;
+
+      case PARSE_FILE_PARSER:
+        doc = htmlSAXParseFile(THIS->input_data->str, NULL, THIS->sax, NULL);
+        break;
+  }
+  if ( doc != NULL )
+    xmlFreeDoc(doc);
   
   push_int(0);
 }
@@ -1108,6 +1290,17 @@ static void init_sax(struct object *o)
   econtent_prefix = make_shared_string("prefix");
   econtent_c1 = make_shared_string("child1");
   econtent_c2 = make_shared_string("child2");
+  seDomain = make_shared_string("domain");
+  seCode = make_shared_string("code");
+  seMessage = make_shared_string("message");
+  seLevel = make_shared_string("level");
+  seFile = make_shared_string("file");
+  seLine = make_shared_string("line");
+  seStr1 = make_shared_string("str1");
+  seStr2 = make_shared_string("str2");
+  seStr3 = make_shared_string("str3");
+  seInt1 = make_shared_string("int1");
+  seInt2 = make_shared_string("int2");
 }
 
 static void exit_sax(struct object *o)
@@ -1124,6 +1317,17 @@ static void exit_sax(struct object *o)
   if (econtent_prefix) free_string(econtent_prefix);
   if (econtent_c1) free_string(econtent_c1);
   if (econtent_c2) free_string(econtent_c2);
+  if (seDomain) free_string(seDomain);
+  if (seCode) free_string(seCode);
+  if (seMessage) free_string(seMessage);
+  if (seLevel) free_string(seLevel);
+  if (seFile) free_string(seFile);
+  if (seLine) free_string(seLine);
+  if (seStr1) free_string(seStr1);
+  if (seStr2) free_string(seStr2);
+  if (seStr3) free_string(seStr3);
+  if (seInt1) free_string(seInt1);
+  if (seInt2) free_string(seInt2);
 }
 
 int _init_xml_sax(void)
@@ -1136,13 +1340,28 @@ int _init_xml_sax(void)
 
   ADD_FUNCTION("create", f_create,
                tFunc(tOr(tString, tObj) tObj tOr(tMapping, tVoid) tOr(tMixed, tVoid) tOr(tInt, tVoid), tVoid), 0);
-  ADD_FUNCTION("parse", f_parse, tFunc(tVoid, tInt), 0);
+  ADD_FUNCTION("parse", f_parse_xml, tFunc(tVoid, tInt), 0);
   ADD_FUNCTION("getLineNumber", f_getLineNumber, tFunc(tVoid, tInt), 0);
   ADD_FUNCTION("getColumnNumber", f_getColumnNumber, tFunc(tVoid, tInt), 0);
   
-  sax_program = end_program();
-  add_program_constant("SAX", sax_program, 0);
+  xml_program = end_program();
+  add_program_constant("SAX", xml_program, 0);
 
+  start_new_program();
+  ADD_STORAGE(sax_storage);
+  
+  set_init_callback(init_sax);
+  set_exit_callback(exit_sax);
+
+  ADD_FUNCTION("create", f_create,
+               tFunc(tOr(tString, tObj) tObj tOr(tMapping, tVoid) tOr(tMixed, tVoid) tOr(tInt, tVoid), tVoid), 0);
+  ADD_FUNCTION("parse", f_parse_html, tFunc(tVoid, tInt), 0);
+  ADD_FUNCTION("getLineNumber", f_getLineNumber, tFunc(tVoid, tInt), 0);
+  ADD_FUNCTION("getColumnNumber", f_getColumnNumber, tFunc(tVoid, tInt), 0);
+  
+  html_program = end_program();
+  add_program_constant("HTML", html_program, 0);
+  
   return 1;
 }
 
