@@ -1,6 +1,6 @@
 /*
  * Pike Extension Modules - A collection of modules for the Pike Language
- * Copyright  2000-2003 The Caudium Group
+ * Copyright  2000-2005 The Caudium Group
  * 
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -21,7 +21,7 @@
 /* This file should be included by ALL source code. It includes various
  * stuff to maintain source code compatibility between Pike versions.
  *
- * $Id: caudium_util.h,v 1.9 2003/02/03 17:25:33 kiwi Exp $
+ * $Id: caudium_util.h,v 1.13 2005/01/03 17:08:05 kiwi Exp $
  */
 
 #ifndef HAVE_CAUDIUM_UTIL_H
@@ -48,6 +48,9 @@
 
 #if (PIKE_MAJOR_VERSION == 7 && PIKE_MINOR_VERSION == 1 && PIKE_BUILD_VERSION >= 12) || PIKE_MAJOR_VERSION > 7 || (PIKE_MAJOR_VERSION == 7 && PIKE_MINOR_VERSION > 1)
 # include "pike_error.h"
+#  ifdef fp
+#   undef fp
+#  endif
 #else
 # include "error.h"
 # ifndef Pike_error
@@ -56,10 +59,10 @@
 #endif
 
 /* Pexts version */
-#define PEXTS_VERSION "0.1.1"
+#define PEXTS_VERSION "0.2.0"
 #define PEXTS_MAJOR   0
-#define PEXTS_MINOR   1
-#define PEXTS_BUILD   1
+#define PEXTS_MINOR   2
+#define PEXTS_BUILD   0
 
 #ifndef MODULE_MAJOR
 #define MODULE_MAJOR PEXTS_MAJOR
@@ -163,4 +166,24 @@ pexts_init(void)
 #define ARG(_n_) Pike_sp[-((args - _n_) + 1)]
 #endif
 
+/* Pike compatibility stuff */
+#if (PIKE_MAJOR_VERSION == 7 && PIKE_MINOR_VERSION < 4) || PIKE_MAJOR_VERSION < 7
+#ifndef move_svalue
+#define move_svalue(TO, FROM) do {					\
+    struct svalue *_to = (TO);						\
+    struct svalue *_from = (FROM);					\
+    dmalloc_touch_svalue(_from);					\
+    *_to = *_from;							\
+  } while (0)
+#endif
+
+#ifndef stack_pop_to_no_free
+#define stack_pop_to_no_free(X) move_svalue(X, --Pike_sp)
+#endif
+
+#ifndef stack_pop_to
+#define stack_pop_to(X) do { struct svalue *_=(X); free_svalue(_); stack_pop_to_no_free(_); }while(0)
+#endif
+#endif /* old pike */
+
 #endif /* HAVE_CAUDIUM_UTIL_H */
